import java.util.*;

public class SMBoundedEnv extends SquareEnvironment
{
    // Instance Variables: Encapsulated data for each BoundedEnv object
    	private LinkedList[] matrix;  // grid representing the environment
    	private int objectCount;        // # of objects in current environment
	private int maxRows;
	private int maxCols;

  // constructors

    /** Constructs an empty BoundedEnv object with the given dimensions.
     *  (Precondition: <code>rows > 0</code> and <code>cols > 0</code>.)
     *  @param rows        number of rows in BoundedEnv
     *  @param cols        number of columns in BoundedEnv
     **/
    public SMBoundedEnv(int rows, int cols)
    {
        // Construct and initialize inherited attributes.
        	super();
		maxRows = rows;
		maxCols = cols;
		matrix = new LinkedList[rows];
		for(int j = 0; j < rows; j++)
		    matrix[j] = new LinkedList();
       
        objectCount = 0;
    }


  // accessor methods

    /** Returns number of rows in the environment.
     *  @return   the number of rows, or -1 if this environment is unbounded
     **/
    public int numRows()
    {
        return maxRows;
    }

    /** Returns number of columns in the environment.
     *  @return   the number of columns, or -1 if this environment is unbounded
     **/
    public int numCols()
    {
        // Note: according to the constructor precondition, numRows() > 0, so
        // theGrid[0] is non-null.
        return maxCols;
    }

    /** Verifies whether a location is valid in this environment.
     *  @param  loc    location to check
     *  @return <code>true</code> if <code>loc</code> is valid;
     *          <code>false</code> otherwise
     **/
    public boolean isValid(Location loc)
    {
        if ( loc == null )
            return false;

        return (0 <= loc.row() && loc.row() < numRows()) &&
               (0 <= loc.col() && loc.col() < numCols());
    }

    /** Returns the number of objects in this environment.
     *  @return   the number of objects
     **/
    public int numObjects()
    {
        return objectCount;
    }

    /** Returns all the objects in this environment.
     *  @return    an array of all the environment objects
     **/
    public Locatable[] allObjects()
    {
        Locatable[] theObjects = new Locatable[numObjects()];
        int index = 0;

        // Look at all grid locations.
        for ( int r = 0; r < numRows(); r++ )
        {
         	LinkedList list = matrix[r];
         	Iterator itr = list.iterator();
         	while(itr.hasNext())
         	{
         		theObjects[index] = (Locatable)itr.next();
         		index++;
         	}
        }

        return theObjects;
    }

    /** Determines whether a specific location in this environment is
     *  empty.
     *  @param loc  the location to test
     *  @return     <code>true</code> if <code>loc</code> is a
     *              valid location in the context of this environment
     *              and is empty; <code>false</code> otherwise
     **/
    public boolean isEmpty(Location loc)
    {
        return isValid(loc) && objectAt(loc) == null;
    }

    /** Returns the object at a specific location in this environment.
     *  @param loc    the location in which to look
     *  @return       the object at location <code>loc</code>;
     *                <code>null</code> if <code>loc</code> is not
     *                in the environment or is empty
     **/
    public Locatable objectAt(Location loc)
    {
        if ( ! isValid(loc) )
            return null;

        LinkedList list = matrix[loc.row()];
        Iterator itr = list.iterator();
        while(itr.hasNext())
        {
        	Locatable fish = (Locatable)itr.next();
        	if(fish.location().equals(loc))
        	{
        		return fish;
        	}
        }
        return null;
    }

    /** Creates a single string representing all the objects in this
     *  environment (not necessarily in any particular order).
     *  @return    a string indicating all the objects in this environment
     **/
    public String toString()
    {
        Locatable[] theObjects = allObjects();
        String s = "Environment contains " + numObjects() + " objects: ";
        for ( int index = 0; index < theObjects.length; index++ )
            s += theObjects[index].toString() + " ";
        return s;
    }


  // modifier methods

    /** Adds a new object to this environment at the location it specifies.
     *  (Precondition: <code>obj.location()</code> is a valid empty location.)
     *  @param obj the new object to be added
     *  @throws    IllegalArgumentException if the precondition is not met
     **/
    public void add(Locatable obj)
    {
        // Check precondition.  Location should be empty.
        Location loc = obj.location();
        if ( ! isEmpty(loc) )
            throw new IllegalArgumentException("Location " + loc +
                                    " is not a valid empty location");

        // Add object to the environment.
        matrix[obj.location().row()].add(obj);
        objectCount++;
    }

    /** Removes the object from this environment.
     *  (Precondition: <code>obj</code> is in this environment.)
     *  @param obj     the object to be removed
     *  @throws    IllegalArgumentException if the precondition is not met
     **/
    public void remove(Locatable obj)
    {
        // Make sure that the object is there to remove.
        Location loc = obj.location();
        if ( objectAt(loc) != obj )
            throw new IllegalArgumentException("Cannot remove " + 
                                               obj + "; not there");

        // Remove the object from the grid.
       Iterator itr = matrix[loc.row()].iterator();
       while(itr.hasNext())
       {
       	   Locatable fish = (Locatable)itr.next();
       	   if(obj.equals(fish))
       	   {
       	   		itr.remove();
       	   		objectCount--;
       	   		break;
       	   }
       	}
    }

    /** Updates this environment to reflect the fact that an object moved.
     *  (Precondition: <code>obj.location()</code> is a valid location
     *  and there is no other object there.
     *  Postcondition: <code>obj</code> is at the appropriate location
     *  (<code>obj.location()</code>), and either <code>oldLoc</code> is 
     *  equal to <code>obj.location()</code> (there was no movement) or
     *  <code>oldLoc</code> is empty.)
     *  @param obj       the object that moved
     *  @param oldLoc    the previous location of <code>obj</code>
     *  @throws    IllegalArgumentException if the precondition is not met
     **/
    public void recordMove(Locatable obj, Location oldLoc)
    {
        // Simplest case: There was no movement.
        Location newLoc = obj.location();
        if ( newLoc.equals(oldLoc) )
            return;

        // Otherwise, oldLoc should contain the object that is
        //   moving and the new location should be empty.
        boolean duplicate = false;
        LinkedList list = matrix[oldLoc.row()];
        Iterator itr = list.iterator();
        while(itr.hasNext())
        {
        	Locatable fish = (Locatable)itr.next();
        	if(fish.location().equals(oldLoc) )
                 duplicate = true;
        }
        	     
        
        if ( duplicate)
            throw new IllegalArgumentException("Precondition violation moving "
                + obj + " from " + oldLoc);
                
        boolean found = false;
        itr = list.iterator();
        while(itr.hasNext())
        {
        	if(obj.equals(itr.next()))
        	{
        		itr.remove();
        		found = true;
        		break;
        	}
        }        
      	if(!found && !isEmpty(newLoc))
            throw new IllegalArgumentException("Precondition violation moving "
                + obj + " from " + oldLoc);
        	
        // Move the object to the proper location in the grid.
    
        
        matrix[obj.location().row()].add(obj);
    }

}
